/* Base */

@import "../module/**/*.less";

  html {
    overflow-y: scroll;
  }
  .row {
    margin-left: 0;
    margin-right: 0;
  }
  .row > div {
    margin-bottom: 15px;
  }


.noselect {
    -webkit-touch-callout: none;
    -webkit-user-select: none;
    -khtml-user-select: none;
    -moz-user-select: none;
    -ms-user-select: none;
    user-select: none;
}
