.logcontainer {
    width:100%;
    padding:0px;
    margin-top:-20px;
    pre {
        margin:0px;
    }
}
pre.log {
    div {
        height: 18px;
    }
    .linenumber {
        display: inline-block;
        width: 50px;
    }
    // It is much more difficult to do autoscroll if line wrap is enabled.
    // For now we just disable it. There is space for some smarter algorithm
    .no-wrap {
        white-space: pre;
        word-wrap: normal;
    }
  margin-top: 0px;
  margin-bottom: 0px;
  overflow: auto;
}

.logoptions {
  margin-top: 0px;
  position: absolute;
  right: 17px;
}
.loading {
  margin-top: -40px;
  position: absolute;
  right: 17px;
  z-index:10000;
}
