input.select-editable {
        /* we put the input on top of the select, with a width of 90%
           so that the right side of the select if visible */
        position:absolute;
        top:0;
        border:none;
        margin:2px;
        width:90%;
        height:29px;
}
