<svg xmlns="http://www.w3.org/2000/svg" width="{{ left.width+config.left_pad + right.width+config.right_pad +2}}" height="18">
  <linearGradient id="smooth" x2="0" y2="100%">
    <stop offset="0"  stop-color="#fff" stop-opacity=".7"/>
    <stop offset=".1" stop-color="#aaa" stop-opacity=".1"/>
    <stop offset=".9" stop-color="#000" stop-opacity=".3"/>
    <stop offset="1"  stop-color="#000" stop-opacity=".5"/>
  </linearGradient>
  <rect rx="{{ config.border_radius }}" width="{{ left.width+config.left_pad + right.width+config.right_pad }}" height="18" fill="{{ left.color }}"/>
  <rect rx="{{ config.border_radius }}" x="{{ left.width+config.left_pad +2}}" width="{{ right.width+config.right_pad }}" height="18" fill="{{ right.color }}"/>
  <rect x="{{ left.width+config.left_pad+2 }}" width="4" height="18" fill="{{ right.color }}"/>
  <rect rx="{{ config.border_radius }}" width="{{ left.width+config.left_pad+right.width+config.right_pad+2 }}" height="18" fill="url(#smooth)"/>
  <g fill="#fff" text-anchor="middle" font-family="DejaVu Sans,Verdana,Geneva,sans-serif" font-size="11">
    <text x="{{ (left.width+config.left_pad)/2+1 }}" y="14" fill="#010101" fill-opacity=".3">{{ left.text }}</text>
    <text x="{{ (left.width+config.left_pad)/2+1 }}" y="13">{{ left.text }}</text>
    <text x="{{ (left.width+config.left_pad+2)+(right.width+config.right_pad)/2-1 }}" y="14" fill="#010101" fill-opacity=".3">{{ right.text }}</text>
    <text x="{{ (left.width+config.left_pad+2)+(right.width+config.right_pad)/2-1 }}" y="13">{{ right.text }}</text>
  </g>
</svg>
